from tastypie.resources import ModelResource, Resource
from django.conf import settings
from django.http import HttpResponse

def get_server_url():
    """
    Get the Server URL from the django settings.
    """
    return getattr(settings, 'BASE_SERVER_URL', 'http://localhost:8000')
def get_schema_url():
    """
    Get the URL for the schema.
    """
    return getattr(settings, 'SCHEMA_URL', 'http://localhost:8000/api/v1/schemas/')

class GDAPIisms(object):
    def alter_list_data_to_serialize(self, request, bundle):
        """
        This method alters the main dataset of the response list object.
        Should add a links section.
        """
        if isinstance(bundle, dict):
            if 'meta' in bundle:
                bundle['pagination'] = bundle['meta']
                del(bundle['meta'])
                bundle['type'] = 'collection'
                bundle['resource_type'] = self._meta.resource_name
                bundle['links'] = {}
                bundle['links']['self'] = get_server_url() + request.path

        bundle = self.resource_alter_list_data_to_serialize(request,bundle)

        return bundle

    def resource_alter_list_data_to_serialize(self,request,bundle):
        """
        This method should be implemented ina subclass to customize 
        the response of a collection of jobs. It will run AFTER the
        alter_list_data_to_serialize method in the super 
        class.
        """
        return bundle

    def dehydrate(self,bundle):
        if 'resource_uri' in bundle.data:
            bundle.data = self._create_links_add_self(bundle.data)

        if bundle.data.get('id') and not isinstance(bundle.data['id'],str):
            bundle.data['id']=str(bundle.data['id'])

        bundle = self.resource_dehydrate(bundle)

        return bundle

    def resource_dehydrate(self,bundle):
        """
        This Method should be implemented in a subclass to customize
        the behavior of the dehydrate method. It will run AFTER the
        dehydrate method in the Super class.
        """
        return bundle

    def build_schema(self):
        """
        Returns a dictionary of all the fields on the resource and some
        properties about those fields.

        Used by the ``schema/`` endpoint to describe what will be available.
        """

        server = get_server_url()
        base_url = '{0}/api/v1/'.format(server)
        data = {
            'resource_fields': {},
            'default_format': self._meta.default_format,
            'collection_methods': self._meta.list_allowed_methods,
            'resource_methods': self._meta.detail_allowed_methods,
            'default_limit': self._meta.limit,
            'type': 'schema',
            'id': self._meta.resource_name,
            'links': {
                'self': '{0}/schema/{1}/'.format(base_url, self._meta.resource_name),
                'collection': '{0}{1}/'.format(base_url, self._meta.resource_name)
             }
        }

        data['collection_methods'] = map(lambda x: x.upper(), data['collection_methods'])
        data['resource_methods'] = map(lambda x: x.upper(), data['resource_methods'])

        if self._meta.ordering:
            data['ordering'] = self._meta.ordering

        if self._meta.filtering:
            data['filtering'] = self._meta.filtering

        for field_name, field_object in self.fields.items():
            data['resource_fields'][field_name] = {
                'default': field_object.default,
                'type': field_object.dehydrated_type,
                'nullable': field_object.null,
                'create': not field_object.readonly,
                'update': not field_object.readonly,
                'unique': field_object.unique,
                'required': True if not field_object.null and not field_object.blank and field_object.default == None else False,
                'valid_chars': None,
                'invalid_chars': None,
                'min': None,
                'max': None,
                'min_length': None,
                'max_length': None,
                'options': None,
            }
            if field_object.dehydrated_type == 'related':
                if getattr(field_object, 'is_m2m', False):
                    related_type = 'array[reference[{0}]]'.format(field_object.to_class()._meta.resource_name)
                else:
                    related_type = 'reference[{0}]'.format(field_object.to_class()._meta.resource_name)

                data['resource_fields'][field_name]['type'] = related_type

        del(data['resource_fields']['resource_uri'])

        data = self.resource_build_schema(data)
       
        return data

    def resource_build_schema(self, data):
        """
        Implement this at the resource level. This handles the majority
        Of the tweaking.
        """
        return data

    def _create_links_add_self(self, dictionary):
        if 'resource_uri' in dictionary:
            dictionary['links'] = {}
            dictionary['links']['self'] = dictionary['resource_uri']
            del(dictionary['resource_uri'])

        dictionary = self._resource_add_links(dictionary)

        return dictionary

    def _resource_add_links(self, dictionary):
        """
        This method should be implemented in a subclass to add
        links to the 'links' key.
        """

        return dictionary

class GDHttpResponse(HttpResponse):
    """
    Overriding to add a default header...
    """
    def __init__(self, content=b'', *args, **kwargs):
        super(GDHttpResponse, self).__init__(content, *args, **kwargs)
        self['X-API-Schema'] = get_schema_url()

class GDApiModelResource(GDAPIisms,ModelResource):
    """
    This class should be inherited, and not used directly. It does
    some of the basic manipulations to match the GD-API spec. Some
    operations will need to be done at the resource level.
    """
    def create_response(self, request, data, response_class=GDHttpResponse, **response_kwargs):
        return super(GDApiModelResource, self).create_response(request, data, response_class, **response_kwargs)

    def get_resource_uri(self, bundle_or_obj=None, url_name='api_dispatch_list'):
        path = super(GDApiModelResource, self).get_resource_uri(bundle_or_obj, url_name)
        server_url = get_server_url()

        return server_url + path

class GDApiResource(GDAPIisms, Resource):
    """
    This class should be inherited, and not used directly. It does
    some of the basic manipulations to match the GD-API spec. Some
    operations will need to be done at the resource level.
    """
    def create_response(self, request, data, response_class=GDHttpResponse, **response_kwargs):
        return super(GDApiResource, self).create_response(request, data, response_class, **response_kwargs)

    def get_resource_uri(self, bundle_or_obj=None, url_name='api_dispatch_list'):
        path = super(GDApiResource, self).get_resource_uri(bundle_or_obj, url_name)
        server_url = get_server_url()

        return server_url + path

