"""Provides Authentication and Authorization based on the Go Daddy Auth service.

The GDAuthentication object can be used to check if a request has a
GDAuthToken specified in the auth_idp cookie.  If the cookie exists
and is validated with the GDAuth service then the user is
authenticated.

Once authenticated (via the GDAuthentication or other means), the
GDShopperIdAuthorization can be used as an Authorization element for
operations.  The base GDShopperIdAuthorization element allows access
for RU operations (viz. CRUD database operations) but denies all CD
operations (since those should probably be initiated via billing).

As a small deviation from TastyPie defaults, these elements will raise
Unauthorized (401) if the cookie does not exist or is invalid, but
they will raise a 403 error if the cookie exists but is not valid for
the object requested.  This allows callers to redirect to a login on a
401, but a 403 is not resolveable by a login.

This module also defines a CascadeAuthentication and
CascadeAuthorization class.  These can be used as the primary Auth
elements; explicit Authentication/Authorization instances are
registered with them and they will try each one in sequence.  Any True
is immediately accepted, a 401 or False will propagate to check the
next registered instance, but a 403 (or other exception) is returned
immediately.

"""

from tastypie.authentication import Authentication
from tastypie.authorization import Authorization
from tastypie.exceptions import Unauthorized, ImmediateHttpResponse
from tastypie.http import HttpForbidden
from django.conf import settings
from .gdauthtoken import GDAuthToken, GDEmployeeToken
import logging



class GDEmployeeAuthentication(Authentication):
    "Authenticate an employee that has an auth_jomax GDAuthToken cookie."

    authToken = lambda s,cookie: GDEmployeeToken(cookie,
                                                 getattr(settings, 'GDAUTH_URL',
                                                         'https://sso.godaddy.com/v1/api'))

    def is_authenticated(self, request, **kw):
        """Return True if request has an auth_idp cookie that is a validatable
           GDAuthToken, otherwise False for no cookie or invalid
           cookie.
        """
        logging.debug("entering GDEmployeeAuthentication.is_authenticated")
        if 'auth_jomax' in request.COOKIES:
            logging.debug("GDEmployeeAuthentication: entering is_authenticated @ auth_jomax")
            try:
                request.gdauthtoken = self.authToken(request.COOKIES['auth_jomax'])

                # Exception thrown above if cookie is invalid/expired
                return True
            except ValueError as ex:
                # cookie is not valid.  Authentication REJECTED.
                logging.getLogger('GD.Auth').warning('Auth Cookie Error: %s', str(ex))
                return False

        return False

    class EmptyRequest:
        '''used to fake out the get_identifier method's get attribute default
        return value.  if the request passed doesn't have the gdauthtoken attribute
        this class will be returned for firstname to be executed on.'''
        def accountName(self):
            return 'SOME RANDOM EMPLOYEE'

    def get_identifier(self, request):
        return getattr(request, 'gdauthtoken', self.EmptyRequest()).accountName()


class GDAuthentication(Authentication):
    "Authenticate a customer that has an auth_idp GDAuthToken cookie."

    authToken = lambda s,cookie: GDAuthToken(cookie, getattr(settings, 'GDAUTH_URL',
                                                             'https://sso.godaddy.com/v1/api'))

    def is_authenticated(self, request, **kw):
        """Return True if request has an auth_idp cookie that is a validatable
           GDAuthToken, otherwise False for no cookie or invalid
           cookie.
        """
        if 'auth_idp' in request.COOKIES:
            try:
                request.gdauthtoken = self.authToken(request.COOKIES['auth_idp'])
                # Exception thrown above if cookie is invalid/expired
                return True
            except ValueError as ex:
                # cookie is not valid.  Authentication REJECTED.
                logging.getLogger('GD.Auth').warning('Auth Cookie Error: %s', str(ex))
                return False
        return False

    class EmptyRequest:
        '''used to fake out the get_identifier method's get attribute default
        return value.  if the request passed doesn't have the gdauthtoken attribute
        this class will be returned for firstname to be executed on.'''
        def firstname(self):
            return 'SOME RANDOM PERSON'

    def get_identifier(self, request):
        return getattr(request, 'gdauthtoken', self.EmptyRequest()).firstname()

class GDApiAuthentication(Authentication):
    authToken = lambda s,cookie: GDAuthToken(cookie, getattr(settings, 'GDAUTH_URL',
                                                             'https://sso.godaddy.com/v1/api'))

    def is_authenticated(self, request, **kw):
        """Return True if request has an idp-jwt header that is a validatable
           GDAuthToken, otherwise False for no cookie or invalid
           cookie.
        """
        token = self.get_token(request)
        if token is not None:
            try:
                request.gdauthtoken = self.authToken(token)
                # Exception thrown above if cookie is invalid/expired
                return True
            except ValueError as ex:
                # cookie is not valid.  Authentication REJECTED.
                logging.getLogger('GD.Auth').warning('Auth Cookie Error: %s', str(ex))
                return False
        return False

    def get_token(self, request):
        if request.META.get('HTTP_AUTHORIZATION') and request.META['HTTP_AUTHORIZATION']:
            (auth_type, token) = request.META['HTTP_AUTHORIZATION'].split()

            if auth_type.lower() != 'idp-jwt':
                raise ValueError("Incorrect authorization header.")

            return token

_oAuthCheckFun  = lambda o: getattr(o, 'authorization_check', lambda shopperId: True)


class GDShopperIdAuthorization(Authorization):

    """If the request has a gdauthtoken (see GDAuthentication) then the
    shopperId in that token must match objects to allow read and
    update operations.  No create or delete operations are allowed if
    there is a gdauthtoken because these operations involve billing
    and must come from internal tools (basic authentication, not
    gdauthtoken) instead.

    If there is no gdauthtoken (basic authentication used), or the
    token is a GDEmployeeToken, allow all operation (this assumes that
    this is therefore developer/operational access which is
    (currently) unrestricted).

    """

    def __init__(self, filterkey = 'shopperId', *args, **kw):
        super(GDShopperIdAuthorization, self).__init__(*args, **kw)
        self.filterkey = filterkey

    def read_list(self, object_list, bundle):
        # Assumes object_list is a QuerySet from ModelResource
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if not tok or isinstance(tok, GDEmployeeToken):
            return object_list
        fdict = { self.filterkey: tok.shopperId() }
        return object_list.filter(**fdict)

    def read_detail(self, object_list, bundle):
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if tok and not isinstance(tok, GDEmployeeToken) and \
           not _oAuthCheckFun(bundle.obj)(shopperId=tok.shopperId()):
            try:
                logging.getLogger('GD.Auth').warning('Read Authorization Failure'
                                                     ', %s %s, shopperId %s',
                                                     str(bundle.obj.__class__.__name__),
                                                     str(bundle.obj.id),
                                                     str(tok.shopperId()))
            except:
                logging.getLogger('GD.Auth').warning('Read Authorization Failure'
                                                     ', some %s, shopperId %s',
                                                     str(bundle.obj.__class__.__name__),
                                                     str(tok.shopperId()))
            raise ImmediateHttpResponse(HttpForbidden("can't touch this!"))
        return True


    def create_list(self, object_list, bundle):
        # Only internal tools are allowed to create, not shoppers: fail if there's a token.
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if not tok or isinstance(tok, GDEmployeeToken):
            return object_list
        return []

    def create_detail(self, object_list, bundle):
        # Only internal tools are allowed to create, not shoppers: fail if there's a token.
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if not tok or isinstance(tok, GDEmployeeToken):
            return True
        raise ImmediateHttpResponse(HttpForbidden("no simple create of a billable resource"))


    def update_list(self, object_list, bundle):
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if tok and not isinstance(tok, GDEmployeeToken):
            return [obj for obj in object_list
                    if _oAuthCheckFun(obj)(shopperId=tok.shopperId())]
        return object_list

    def update_detail(self, object_list, bundle):
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if tok and not isinstance(tok, GDEmployeeToken) and \
           not _oAuthCheckFun(bundle.obj)(shopperId=tok.shopperId()):
            raise ImmediateHttpResponse(HttpForbidden("untouchable"))
        return True


    def delete_list(self, object_list, bundle):
        # Like create, only internal tools are allowed to delete, not shoppers
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if not tok or isinstance(tok, GDEmployeeToken):
            # Safety check: there must be only one!  The delete_list
            # is needed to support void calling DELETE on
            # "/containers/?accountUid=...", but since we devs use
            # curl, a mistaken DELETE on "/containers/" would
            # disastrously delete EVERYTHING.  Yes, ALL CONTAINERS and
            # the ENTIRE DATABASE would be deleted!!!!
            # So... authorized users can only delete one at a time for
            # safety.
            if len(object_list) == 1: return object_list
            return []
        return []

    def delete_detail(self, object_list, bundle):
        tok = getattr(bundle.request, 'gdauthtoken', None)
        if not tok or isinstance(tok, GDEmployeeToken):
            return True
        raise ImmediateHttpResponse(HttpForbidden("no simple delete of a billable resource"))



class CascadeAuthentication(Authentication):

    "Request is authenticated if any Authenticator recognizes the request settings"

    def __init__(self, *args, **kw):
        super(CascadeAuthentication, self).__init__(*args, **kw)
        self.authenticaters = []

    def addAuthenticater(self, newAuth):
        self.authenticaters.append(newAuth)

    def is_authenticated(self, request, **kw):
        for auth in self.authenticaters:
            # warning: tastypie docs imply is_authenticated should
            # return True or False, but it's possible to return an
            # isinstance of HttpResponse as well (which is immediately
            # returned as the auth failure).  Here therefore we must
            # explicitly test for True for success.
            if auth.is_authenticated(request, **kw) is True:
                logging.debug("entering is_authenticated true")
                if isinstance(auth, GDShopperIdAuthorization):
                    request.cascade_userid = auth.get_identifier(request)
                return True

        return False

    def get_identifier(self, request):
        id = getattr(request, "cascade_userid", None)
        if id is not None:
            return id
        for auth in self.authenticaters:
            if auth.is_authenticated(request) is True:
                return auth.get_identifier(request)
        return '<unidentified>'

class CascadeAuthorization(Authorization):

    """Authorization is granted if any Authorizers allows it.  Object
       filtering is the sum of all filtering.
    """

    def __init__(self, *args, **kw):
        super(CascadeAuthorization, self).__init__(*args, **kw)
        self.authorizers = []

        self.read_list   = self.auth_list("read_list")
        self.read_detail = self.auth_detail("read_detail")

        self.create_list   = self.auth_list("create_list")
        self.create_detail = self.auth_detail("create_detail")

        self.update_list   = self.auth_list("update_list")
        self.update_detail = self.auth_detail("update_detail")

        self.delete_list   = self.auth_list("delete_list")
        self.delete_detail = self.auth_detail("delete_detail")


    def addAuthorizer(self, newAuth):
        self.authorizers.append(newAuth)

    def auth_list(self, auth_op):
        def alf(self, object_list, bundle):
            for auth in self.authorizers:
                object_list = getattr(auth, auth_op)(object_list, bundle)
            return object_list
        return alf.__get__(self, CascadeAuthorization)

    def auth_detail(self, auth_op):
        def adf(self, object_list, bundle):
            for auth in self.authorizers:
                try:
                    if getattr(auth, auth_op)(object_list, bundle):
                        return True
                except Unauthorized:
                    # Assume this is "not accepted" but not
                    # "explicitly rejected".  Anything else is assume
                    # to be explicit rejection and not caught.  N.B.
                    # recommend using `raise
                    # ImmediateHttpResponse(HttpForbidden()) as
                    # explicit rejection.
                    pass
            raise Unauthorized('no valid authorization')
        return adf.__get__(self, CascadeAuthorization)

