"""Provides the GDAuthToken object, which is an object representation of the Authorization Token defined by https://confluence.int.godaddy.com/display/AUTH/Integration+Guide"""

# pip-requirements:
# pycrypto
# requests

from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from Crypto.Hash import SHA256
from base64 import b16encode, urlsafe_b64decode
from django.conf import settings
import datetime
import requests
import json
import sys

def _decode(s):
    return urlsafe_b64decode((s + '=' * (4 - (len(s) % 4))).encode('utf-8'))
_long = int if sys.version_info[0] >= 3 else long #py3: long is gone
def _kconv(arg): return _long(b16encode(_decode(arg)), 16)

class GDToken(object):
    """Given the cookie-version of a GD Auth Token, this object decodes
       and validates the token and provides retrieval methods for
       various elements of the token.  An exception is thrown in the
       contructor if the token is not valid; existence of a
       GDAuthToken implies that it is valid (although it may expire
       while being held).

       For validation, the token signing public key may be retrieved
       from the GD Auth server and used for validation; retrieved
       tokens are cached and therefore only retrieved once.
    """

    def __init__(self, cookieval, gdAuthURL='https://sso.godaddy.com/v1/api'):
        self.gdAuthURL = gdAuthURL
        p = cookieval.split('.')
        if len(p) != 3:
            raise ValueError('Auth Token is not tripartite: %s'%str(cookieval))
        self.sdata = cookieval[:-len(p[2])-1]
        l = [_decode(x) for x in p]
        self.hdr = json.loads(l[0].decode('utf-8'))
        self.dta = json.loads(l[1].decode('utf-8'))
        print("header: %s"% self.hdr)
        print("dta: %s"% self.dta)
        self.sig = l[2]
        self._validate()
        if self.expired():
            raise ValueError('Auth Token is expired')

    def firstname(self): return self.dta['firstname']

    def expired(self):
        return datetime.datetime.utcfromtimestamp(self.dta['exp']) < datetime.datetime.now()

    def _validate(self):
        kid = self.hdr['kid']
        if kid not in self.__class__.keys:
            self.__class__.keys[kid] = self._fetch_auth_pubkey(kid)
        pubkey = self.__class__.keys[kid]
        if not self._key_verify(self.sdata.encode('utf-8'), self.sig, pubkey, self.hdr['alg']):
            raise ValueError('Auth Token invalid')

    keys = {}  # key ID: gdauth public key from web service (global mapping)

    def _fetch_auth_pubkey(self, keyId):
        print(self.gdAuthURL + '/key/' + keyId)
        authResp = requests.get(self.gdAuthURL + '/key/' + keyId)
        if authResp.status_code != 200:
            raise ValueError('Auth Server status %d'%authResp.status_code)
        rval = dict([(v, _kconv(authResp.json()['data'][v])) for v in ['e', 'n']])
        return rval

    @staticmethod
    def _key_verify(signedData, signature, pubkey, alg):
        pKey = RSA.construct( (pubkey['n'], pubkey['e']) )
        return PKCS1_v1_5.new(pKey).verify(SHA256.new(signedData), signature)

class GDAuthToken(GDToken):

    def shopperId(self): return self.dta['shopperId']

class GDEmployeeToken(GDToken):

    def accountName(self): return self.dta['accountName']
    def groups(self): return self.dta['groups']

